-- actions based on chains of linked entities.
-- used MaxRateCalculator, RateCalculator and others to learn how to make a selector.

log("Initializing Chain Action.")

local function pn(message) -- will only print duplicated messages once
	game.print(message, {255, 255, 255})
	log(message)
end

local function pne(entity)
	pn("entity " .. entity.name .. " type " .. entity.type .. " direction " .. entity.direction .. " " .. serpent.line(entity.position))
end
-- local function pne(count, entity)
-- 	pn(count.."entity " .. entity.name .. " type " .. entity.type .. serpent.line(entity.position))
-- end

local function chain_action(player, tool_name)
	local old_hand = ""
	if player.cursor_stack ~= nil and player.cursor_stack.valid_for_read
	then
        old_hand = player.cursor_stack.name
	end 
	player.clear_cursor()
    if player.cursor_stack ~= nil and old_hand ~= tool_name -- if already in hand, just clear it and get out
    then
        player.cursor_stack.clear()
        player.cursor_stack.set_stack({name=tool_name, type="selection-tool", count = 1})
	end
end

local function chain_action_decon_key(event)
	local player = game.players[event.player_index]
    chain_action(player, "chain-action-decon-tool")
end

local function chain_action_upgrade_key(event)
	local player = game.players[event.player_index]
    chain_action(player, "chain-action-upgrade-tool")
end

local function entity_facing(entity, other_entity)
	-- west is x-1, south = y+1
	local delta_x = entity.position.x - other_entity.position.x
	local delta_y = entity.position.y - other_entity.position.y
	
	if(   other_entity.direction == defines.direction.north      and delta_x ==  0 and delta_y == -1) -- N
	then
		return true
	elseif(other_entity.direction == defines.direction.northeast and delta_x ==  1 and delta_y == -1) -- NE
	then
		return true
	elseif(other_entity.direction == defines.direction.east      and delta_x ==  1 and delta_y ==  0) -- E
	then
		return true
	elseif(other_entity.direction == defines.direction.southeast and delta_x ==  1 and delta_y ==  1) -- SE
	then
		return true
	elseif(other_entity.direction == defines.direction.south     and delta_x ==  0 and delta_y ==  1) -- S
	then
		return true
	elseif(other_entity.direction == defines.direction.southwest and delta_x == -1 and delta_y ==  1) -- SW
	then
		return true
	elseif(other_entity.direction == defines.direction.west      and delta_x == -1 and delta_y ==  0) -- W
	then
		return true
	elseif(other_entity.direction == defines.direction.northwest and delta_x == -1 and delta_y == -1) -- NW
	then
		return true
	else
		return false
	end
end

-- local belts = {"transport-belt", "splitter", "underground-belt"} -- all types of belt
local belts = {"transport-belt"}
-- local pipes = {"pipe", "pipe-to-ground"}
local pipes = {"pipe"}
local list = {}

local function is_type(entity, type)
	for _, v in pairs(type)
	do
		if(entity.name == "entity-ghost")
		then
			if(entity.ghost_type == v)
			then
				return true
			end
		else
			if(entity.type == v)
			then
				return true
			end
		end
	end
	return false
end

local function list_insert(entity)
	if(entity ~= nil)
	then
		list[entity.name ..":".. entity.position.x ..":".. entity.position.y] = entity
	end
end

local function in_list(entity)
	if(list[entity.name ..":".. entity.position.x ..":".. entity.position.y] ~= nil)
	then
		return true
	else
		return false
	end
end

local function find_similar_belt(entity)
	if(entity.type == "entity-ghost")
	then
		pn("ghost")
		-- ghost belt_neighbours doesn't work, so use direction finding...
		-- entity directions: 0 = north, 2 = east, 4 = south, 6 = west   should probably use some global constant for this... defines.direction
		local found = entity.surface.find_entities_filtered({position=entity.position, radius=1.5, type=belts}) -- need to use a set of types here, so we get belts/splitters/undergrounds for belts, etc
		-- pn("real ".. serpent.block(found))
		local foundg = entity.surface.find_entities_filtered({position=entity.position, radius=1.5, ghost_type=belts}) -- need to use a set of types here, so we get belts/splitters/undergrounds for belts, etc
		-- pn("ghost" .. serpent.block(found))
		-- for some reason, it finds splitters if they are in the list, but doesn't include them below int he action, so removed them for now...
		for _, entity2 in pairs(found)
		do
			if(entity2 ~= nil)
			then
				pn("real " .. entity2.name)
				-- found an nearby entity, see if it is facing given entity
				if(entity_facing(entity, entity2) and not in_list(entity2)) -- facing current belt and not already noted, add to list and recurse
				then
					list_insert(entity2)
					find_similar_belt(entity2)
				elseif(entity_facing(entity2, entity) and not in_list(entity2)) -- current belt facing new belt and not already noted, add to list and recurse
				then
					list_insert(entity2)
					find_similar_belt(entity2)
				end
			end
		end	
		for _, entity2 in pairs(foundg)
		do
			if(entity2 ~= nil)
			then
				pn("ghost " .. entity2.name)
				-- found an nearby entity, see if it is facing given entity
				if(entity_facing(entity, entity2) and not in_list(entity2)) -- facing current belt and not already noted, add to list and recurse
				then
					list_insert(entity2)
					find_similar_belt(entity2)
				elseif(entity_facing(entity2, entity) and not in_list(entity2)) -- current belt facing new belt and not already noted, add to list and recurse
				then
					list_insert(entity2)
					find_similar_belt(entity2)
				end
			end
		end	
		return
	end
	pn("type " .. entity.type)
	pn("inputs")
	if(entity.belt_neighbours.inputs ~= nil)
	then
		for _, entity2 in pairs(entity.belt_neighbours.inputs)
		do
			pne(entity2)
			if(not in_list(entity2) and is_type(entity2, belts))
			then
				list_insert(entity2)
				find_similar_belt(entity2)
			end		
		end	
	end
	pn("outputs")
	if(entity.belt_neighbours.outputs ~= nil)
	then
		for _, entity2 in pairs(entity.belt_neighbours.outputs)
		do
			pne(entity2)
			if(not in_list(entity2) and is_type(entity2, belts))
			then
				list_insert(entity2)
				find_similar_belt(entity2)
			end		
		end	
	end
end

local function find_similar_pipe(entity)
	found, _, _ = entity.neighbours
	for _, entity2 in pairs (found[1])
	do
		if(entity2 ~= nil and not entity2.to_be_deconstructed() and not in_list(entity2) and is_type(entity2, pipes))
		then
			-- found an nearby pipe entity not already marked for deconstruction, so we know its connected...
			-- pne(entity2)
			list_insert(entity2)
			find_similar_pipe(entity2)
		end
	end
end

local function on_selection(event)
	-- pn("on_selection function start")
	if event.item ~= "chain-action-decon-tool" and event.item ~= "chain-action-upgrade-tool"
	then
		return
	end

	-- for k, v in ipairs(event.entities)
	-- do
	-- 	pn("key"..k.."value"..serpent.block(v))
	-- end

	-- pn("number of entities: " .. #event.entities)
	if(#event.entities == 1)
	then
		-- one item selected, find connected items
		local entity = event.entities[1]
		if(event.item == "chain-action-upgrade-tool" and entity.prototype.next_upgrade == nil) -- quick check for upgrade tool on non-upgradable item
		then
			return
		end

		list = {}
		-- pne(entity)

		if(is_type(entity, belts))
		then
			list_insert(entity)
			find_similar_belt(entity)
		elseif(is_type(entity, pipes))
		then
			list_insert(entity)
			find_similar_pipe(entity)
		end
	end
	-- pn("list")
	for _, entity in pairs(list)
	do
		if(event.item == "chain-action-decon-tool")
		then
			entity.order_deconstruction(entity.force) -- deconstruct selected object
		elseif(event.item == "chain-action-upgrade-tool")
		then
			entity.order_upgrade({force=entity.force, target=entity.prototype.next_upgrade.name}) -- maybe this works?
		end
		-- pne(entity)
	end
	list = {}
	-- pn("on_selection function end")
end

-- local function on_alt_selection(event)
-- 	if event.item ~= "chain-action-decon-tool" and event.item ~= "chain-action-upgrade-tool"
-- 	then
-- 		return
-- 	end

-- 	if(#event.entities == 1)
-- 	then
-- 		-- one item selected, find connected items
-- 		local entity = event.entities[1]
-- 		-- a, b, c = entity.neighbours
-- 		-- pn(serpent.dump(a)) -- pipes seem to use this one
-- 		-- pn(serpent.dump(b))
-- 		-- pn(serpent.dump(c))
-- 		-- if(a ~= nil)
-- 		-- then
-- 			-- for k, v in ipairs(a)
-- 			-- do
-- 			-- 	pn("key"..k.."value"..serpent.block(v))
-- 			-- end
-- 			-- if(#a == 1)
-- 			-- then
-- 				-- local result = game.player.selected.belt_neighbours
-- 				local result = entity.belt_neighbours
-- 				for _, belt in pairs(result.inputs) do
-- 					rendering.draw_circle{surface=belt.surface, target=belt, color={r=1,g=1,b=0,a=1}, radius=0.4, width=5, time_to_live=300}
-- 				end
-- 				for _, belt in pairs(result.outputs) do
-- 					rendering.draw_circle{surface=belt.surface, target=belt, color={r=1,g=0,b=0,a=1}, radius=0.4, width=5, time_to_live=300}
-- 				end

-- 				-- pn(#a[1])
-- 				-- pn(serpent.dump(a[1]))
-- 				-- for _, ent in pairs(a[1])
-- 				-- do
-- 				-- 	if(ent ~= nil)
-- 				-- 	then
-- 				-- 		pne(ent)
-- 				-- 	end
-- 				-- end
-- 			-- end
-- 		-- end
-- 	end
-- end
script.on_event("chain-action-decon", chain_action_decon_key)
script.on_event("chain-action-upgrade", chain_action_upgrade_key)
script.on_event(defines.events.on_player_selected_area, on_selection)
-- script.on_event(defines.events.on_player_alt_selected_area, on_alt_selection)

-- script.on_event(defines.events.on_player_setup_blueprint, function(event)
--     pn(serpent.dump(event))
-- end)

